const spawn = require('child_process').spawn
const axios = require('axios')
const cmdEscape = require('@ff0000-ad-tech/cmd-escape')

const log = require('@ff0000-ad-tech/debug')('build-source-automator:cs-manager')

function compile(config, servePath, size, verbose) {
	log(``)
	log(`Compiling Test Unit:`)
	return new Promise((resolve, reject) => {
		let cs
		launchCreativeServer(config, servePath, verbose)
			.then(_cs => {
				cs = _cs
				// request webpack launch-command from Creative Server API
				return requestCompileCmd(cs, size)
			})
			.then(cmd => executeCompileCmd(cs, cmd, verbose))
			.then(() => stopCreativeServer(cs))
			.then(() => resolve())
			.catch(err => reject(err))
	})
}

// observe/timeout creative server launch
let originMatch, apiMatch, appMatch
let isLaunched = false
//
const timeoutCsServer = (output, reject) => {
	const timeout = 20 // seconds
	setTimeout(() => {
		if (!isLaunched) {
			reject(new Error(`Creative Server timeout\n${output}`))
		}
	}, timeout * 1000)
}

const checkCsServer = (pid, data, resolve) => {
	const om = data.match(/Origin: (\S+)/)
	if (om && !originMatch) {
		originMatch = om
	}
	const apim = data.match(/API: (\S+)/)
	if (apim && !apiMatch) {
		apiMatch = apim
	}
	const am = data.match(/Server running at (\S+)/)
	if (am && !appMatch) {
		appMatch = am
	}
	if (originMatch && apiMatch && appMatch && !isLaunched) {
		isLaunched = true
		log(` origin: ${originMatch[1]}`)
		log(` api: ${apiMatch[1]}`)
		log(` app: ${appMatch[1]}`)
		resolve({
			pid: pid,
			origin: originMatch[1],
			api: apiMatch[1],
			app: appMatch[1]
		})
	}
}

// launch creative server
function launchCreativeServer(config, servePath, verbose) {
	log(`Launching Creative Server`)
	return new Promise((resolve, reject) => {
		let output = ''

		// command & args
		const cwd = config.cwd || servePath
		const cmd = 'node'
		const cmdArgs = [`${config.wpCsPath}/index.js`]
		const args = ['--context', servePath, '--cwd', cwd, '--wpDmPath', config.wpDmPath, '--timeout', '600'] // if the compile doesn't happen in 10 minutes, kill creative server
		log(` CLI: cd "${cwd}" && ${cmd} ${cmdArgs.join(' ')} ${cmdEscape(args)}`)

		// creative server output
		let env = Object.create(process.env)
		if (!verbose) {
			env.DEBUG = 'wp-creative-server' // prevent express from beltching too much std
		}
		// launch the server
		timeoutCsServer(output, reject)
		const prc = spawn(cmd, cmdArgs.concat(args), { cwd, env })
		prc.on('error', err => {
			reject(err)
		})
		// scan the output for the server-origin
		prc.stdout.on('data', buffer => {
			const data = buffer.toString('utf8')
			if (verbose) {
				log(data)
			}
			output += data
			checkCsServer(prc.pid, data, resolve)
		})
		prc.stderr.on('data', buffer => {
			const data = buffer.toString('utf8')
			if (verbose) {
				log(data)
			}
			output += data
			checkCsServer(prc.pid, data, resolve)
		})
	})
}
const stopCreativeServer = async cs => {
	log(`Stopping Creative Server`)
	const req = `${cs.origin}/api/exit`
	log(` ${req}`)
	await axios.get(req)
}

// curl Creative-Server API for compile command
const requestCompileCmd = async (cs, size) => {
	log(`Requesting compile command`)
	const req = `${cs.origin}/api/get-wp-cmd/default/${size}/index.html`
	log(` ${req}`)
	const res = await axios.get(req)
	return res.data
}

function executeCompileCmd(cs, cmd, verbose) {
	log(``)
	log(`Executing compile command`)
	return new Promise((resolve, reject) => {
		log(` CLI: ${cmd.shell}`)
		let output = ''
		const prc = spawn(cmd.command, cmd.args, { cwd: cmd.cwd })
		prc.stdout.on('data', data => {
			const stdout = data.toString('utf8')
			if (verbose) {
				log(stdout)
			}
			output += stdout
		})
		prc.stderr.on('data', data => {
			const stderr = data.toString('utf8')
			if (verbose) {
				log(stderr)
			}
			output += stderr
		})
		prc.on('error', err => {
			stopCreativeServer(cs)
				.then(() => reject(err))
				.catch(() => reject(err))
		})
		prc.on('close', code => {
			if (code === 0) {
				log(`   - compile complete`)
				resolve()
			} else {
				const err = new Error(`Compile error:\n${output}`)
				stopCreativeServer(cs)
					.then(() => reject(err))
					.catch(() => reject(err))
			}
		})
	})
}

module.exports = {
	compile,
	launchCreativeServer
}
